#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import gzip
import os
import pickle
import tempfile


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def pickle_this(outdir: str, identifier: str, data):
    filename = os.path.join(outdir, identifier + ".pickled.gz")
    with gzip.open(filename, "wb", compresslevel=6) as cpfh:
        pickle.dump(data, cpfh)
    return filename


def unpickle(filename: str):
    data = None
    with gzip.open(filename, "rb") as cpfh:
        data = pickle.load(cpfh)
    return data


def boundary_mode(mode: str) -> int:
    mode = mode.lower()
    if mode == "exact":
        return 0
    if mode == "qint":
        return 1
    if mode == "tinq":
        return 2
    if mode == "all":
        return 3



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import gzip
import json
import logging
import os
import shutil
from time import time
from typing import Dict
from typing import Final
from typing import List
from typing import Optional
from typing import Tuple

import pysam
from pygas.alignercpu import AlignerCpu
from pygas.classes import AlignmentBatch
from pygas.classes import Backtrack

import pycroquet.tools as ctools
from pycroquet import cli
from pycroquet import libparser
from pycroquet import readparser
from pycroquet.classes import Classification
from pycroquet.classes import Library
from pycroquet.classes import Stats
from pycroquet.htscomm import hts_sort_n_index
from pycroquet.main import map_reads
from pycroquet.main import sg_select_alignment
from pycroquet.readwriter import guide_header
from pycroquet.readwriter import read_iter
from pycroquet.readwriter import to_alignment
from pycroquet.readwriter import to_mapped_reads

CLASSIFICATION: Final = Classification()


def best_unique_l_mm_r(
    library: Library, bt_set_l: List[Backtrack], bt_set_r: List[Backtrack]
) -> List[Tuple[int, Backtrack, Backtrack]]:
    """
    find the pairing that is most likely to be the real item
    """
    guides_f_r = {}
    bt_l = bt_set_l[0]
    guide_idxs_l = set(library.target_to_guides[bt_l.sm.target])
    good_idx = None
    for bt_r in bt_set_r:
        guide_idxs_r = set(library.target_to_guides[bt_r.sm.target])
        guide_intersect = guide_idxs_l.intersection(guide_idxs_r)
        for gidx in guide_intersect:
            guide = library.guides[gidx]
            if (
                guide.sgrna_seqs[0] == bt_l.sm.target
                and guide.sgrna_seqs[1] == bt_r.sm.target
                and bt_l.sm.reversed is False
                and bt_r.sm.reversed is True
            ):
                guides_f_r[gidx] = (gidx, bt_l, bt_r)
                good_idx = gidx
    if guides_f_r:
        hits = []
        if len(guides_f_r) > 1:
            # logging.error("Multiple guides appear to equally match query pair, guides are:")
            for gidx in guides_f_r:
                hits.append(guides_f_r[gidx])
                # (_, bt_l, bt_r) = guides_f_r[gidx]
                # logging.error(f"Guide: {library.guides[gidx].sgrna_seqs[0]} | {library.guides[gidx].sgrna_seqs[1]}")
                # logging.error(f"5p query: {bt_l.sm.query} (reversed: {bt_l.sm.reversed})")
                # logging.error(f"3p query: {bt_r.sm.query} (reversed: {bt_r.sm.reversed})")
            # raise ValueError("Before reporting please check '--boundary' is set appropriately.")
        else:
            hits.append(guides_f_r[good_idx])
        return [guides_f_r[good_idx]]
    return None


def best_mm_l_unique_r(
    library: Library, bt_set_l: List[Backtrack], bt_set_r: List[Backtrack]
) -> List[Tuple[int, Backtrack, Backtrack]]:
    """
    find the pairing that is most likely to be the real item
    """
    guides_f_r = {}
    bt_r = bt_set_r[0]
    guide_idxs_r = set(library.target_to_guides[bt_r.sm.target])
    good_idx = None
    for bt_l in bt_set_l:
        guide_idxs_l = set(library.target_to_guides[bt_l.sm.target])
        guide_intersect = guide_idxs_r.intersection(guide_idxs_l)
        for gidx in guide_intersect:
            guide = library.guides[gidx]
            if (
                guide.sgrna_seqs[0] == bt_l.sm.target
                and guide.sgrna_seqs[1] == bt_r.sm.target
                and bt_l.sm.reversed is False
                and bt_r.sm.reversed is True
            ):
                guides_f_r[gidx] = (gidx, bt_l, bt_r)
                good_idx = gidx
    if guides_f_r:
        hits = []
        if len(guides_f_r) > 1:
            # logging.error("Multiple guides appear to equally match query pair, guides are:")
            for gidx in guides_f_r:
                hits.append(guides_f_r[gidx])
                # (_, bt_l, bt_r) = guides_f_r[gidx]
                # logging.error(f"Guide: {library.guides[gidx].sgrna_seqs[0]} | {library.guides[gidx].sgrna_seqs[1]}")
                # logging.error(f"5p query: {bt_l.sm.query} (reversed: {bt_l.sm.reversed})")
                # logging.error(f"3p query: {bt_r.sm.query} (reversed: {bt_r.sm.reversed})")
            # raise ValueError("Before reporting please check '--boundary' is set appropriately.")
        else:
            hits.append(guides_f_r[good_idx])
        return [guides_f_r[good_idx]]
    return None


def best_multimatch_set(library: Library, bt_set_l: List[Backtrack], bt_set_r: List[Backtrack]):
    """
    try to identify a single guide pair when one or both ends are multi-map
    """
    guides = {}  # to hold matches
    last_key = None  # makes getting the single entry far lower impact
    for bt_l in bt_set_l:
        guide_idxs_l = set(library.target_to_guides[bt_l.sm.target])
        for bt_r in bt_set_r:
            guide_idxs_r = set(library.target_to_guides[bt_r.sm.target])
            guide_intersect = guide_idxs_l.intersection(guide_idxs_r)
            if len(guide_intersect) == 1:
                gidx = guide_intersect.pop()
                if bt_l.sm.reversed is False and bt_r.sm.reversed is True:
                    guides[gidx] = (gidx, bt_l, bt_r)
                    last_key = gidx
                    # on multi we don't try to deal with everything
    if len(guides) == 1:
        return guides[last_key]
    return None


def intersect_guide_targets(library, tseq_l, tseq_r):
    guide_idxs_l = set(library.target_to_guides[tseq_l])
    guide_idxs_r = set(library.target_to_guides[tseq_r])
    # gives the common guides
    return guide_idxs_l.intersection(guide_idxs_r)


def classify_read_pair(
    map_l: Tuple[str, List[Backtrack]],
    map_r: Tuple[str, List[Backtrack]],
    library: Library,
) -> Tuple[str, Optional[List[int]], Backtrack, Backtrack, str, str]:  #  noqa R701
    """
    Returns:
        classification string (see classes.Classification)
        None|list of gidx
        R1 backtrack
        R2 backtrack
        R1 map type
        R2 map type
    """
    (mtype_l, mdata_l) = map_l
    (mtype_r, mdata_r) = map_r
    if mtype_l == "unique" and mtype_l == mtype_r:
        (bt_l, bt_r) = (mdata_l[0], mdata_r[0])
        gidx = library.guide_by_sgrna_set(bt_l.sm.target, bt_r.sm.target)
        if gidx is not None:
            if bt_l.sm.reversed is False and bt_r.sm.reversed is True:
                return (CLASSIFICATION.match, gidx, bt_l, bt_r, mtype_l, mtype_r)
            # other orientations will be an abberant match as we are in unique/unique
            return (CLASSIFICATION.aberrant_match, None, bt_l, bt_r, mtype_l, mtype_r)
        # what about r1/r2 order swap
        gidx = library.guide_by_sgrna_set(bt_r.sm.target, bt_l.sm.target)
        if gidx is not None:
            return (CLASSIFICATION.aberrant_match, None, bt_l, bt_r, mtype_l, mtype_r)
        # which leaves swap (as both are unique but to different guide sets)
        return (CLASSIFICATION.swap, None, bt_l, bt_r, mtype_l, mtype_r)
    elif mtype_l == "unmapped" and mtype_l == mtype_r:
        return (CLASSIFICATION.no_match, None, None, None, mtype_l, mtype_r)
    elif mtype_l == "multimap" and mtype_l == mtype_r:
        best_pair = best_multimatch_set(library, mdata_l, mdata_r)
        if best_pair:
            # this will impact how we write out alignments
            (gidx, bt_l, bt_r) = best_pair
            gidx = library.guide_by_sgrna_set(bt_l.sm.target, bt_r.sm.target)
            if gidx is not None:
                if bt_l.sm.reversed is False and bt_r.sm.reversed is True:
                    return (CLASSIFICATION.match, gidx, bt_l, bt_r, mtype_l, mtype_r)
                # other orientations will be an abberant match as we are in unique/unique
                return (CLASSIFICATION.aberrant_match, None, bt_l, bt_r, mtype_l, mtype_r)
            # what about r1/r2 order swap
            gidx = library.guide_by_sgrna_set(bt_r.sm.target, bt_l.sm.target)
            if gidx is not None:
                return (CLASSIFICATION.aberrant_match, None, bt_l, bt_r, mtype_l, mtype_r)
        # which leaves ambiguous (as both map to "things")
        return (CLASSIFICATION.ambiguous, None, None, None, mtype_l, mtype_r)
    elif mtype_l == "unique":
        bt_l = mdata_l[0]
        if mtype_r == "unmapped":
            if bt_l.sm.reversed:
                return (CLASSIFICATION.r_open_3p, None, bt_l, None, mtype_l, mtype_r)
            else:
                return (CLASSIFICATION.f_open_3p, None, bt_l, None, mtype_l, mtype_r)
        # then one end multimap
        best_pairs = best_unique_l_mm_r(library, mdata_l, mdata_r)
        # above function checks orientation, if not as expected this goes to *_multi_3p
        if best_pairs:
            (gidx, bt_l, bt_r) = best_pairs[0]
            if len(best_pairs) > 1:
                return (CLASSIFICATION.f_multi_3p, None, bt_l, None, mtype_l, "multimap")
            gidx = library.guide_by_sgrna_set(bt_l.sm.target, bt_r.sm.target)
            if gidx:
                return (CLASSIFICATION.match, gidx, bt_l, bt_r, mtype_l, mtype_r)
            # what about r1/r2 order swap
            gidx = library.guide_by_sgrna_set(bt_r.sm.target, bt_l.sm.target)
            if gidx:
                return (CLASSIFICATION.aberrant_match, gidx, bt_l, bt_r, mtype_l, mtype_r)
        if bt_l.sm.reversed:
            return (CLASSIFICATION.r_multi_3p, None, bt_l, None, mtype_l, mtype_r)
        return (CLASSIFICATION.f_multi_3p, None, bt_l, None, mtype_l, mtype_r)
    elif mtype_r == "unique":
        bt_r = mdata_r[0]
        if mtype_l == "unmapped":
            if bt_r.sm.reversed:
                return (CLASSIFICATION.r_open_5p, None, None, bt_r, mtype_l, mtype_r)
            else:
                return (CLASSIFICATION.f_open_5p, None, None, bt_r, mtype_l, mtype_r)
        # then one end multimap
        best_pairs = best_mm_l_unique_r(library, mdata_l, mdata_r)
        # above function checks orientation, if not as expected this goes to *_multi_5p
        if best_pairs:
            (gidx, bt_l, bt_r) = best_pairs[0]
            if len(best_pairs) > 1:
                return (CLASSIFICATION.f_multi_5p, None, None, bt_r, "multimap", mtype_r)
            gidx = library.guide_by_sgrna_set(bt_l.sm.target, bt_r.sm.target)
            if gidx:
                return (CLASSIFICATION.match, gidx, bt_l, bt_r, mtype_l, mtype_r)
            # what about r1/r2 order swap
            gidx = library.guide_by_sgrna_set(bt_r.sm.target, bt_l.sm.target)
            if gidx:
                return (CLASSIFICATION.aberrant_match, gidx, bt_l, bt_r, mtype_l, mtype_r)
        if bt_r.sm.reversed:
            return (CLASSIFICATION.r_multi_5p, None, None, bt_r, mtype_l, mtype_r)
        return (CLASSIFICATION.f_multi_5p, None, None, bt_r, mtype_l, mtype_r)
    return (CLASSIFICATION.no_match, None, None, None, mtype_l, mtype_r)


def order_hits(hits: List[Backtrack], first_bt: Backtrack):
    if hits is None or first_bt is None:
        return []
    if len(hits) == 1:
        return hits
    reordered = [first_bt]
    for bt in hits:
        if bt.sm.target != first_bt.sm.target:
            reordered.append(bt)
    return reordered


def _init_class_counts():
    counts = {}
    for v in vars(CLASSIFICATION):
        counts[v] = 0
    return counts


def read_pairs_to_guides(
    workspace: str,
    aligned_results: Dict[str, Tuple[str, List[Backtrack]]],
    library: Library,
    seq_file,
    guide_fa,
    header,
    ref_ids,
    default_rgid,
    stats: Stats,
    cpus=1,
    trim_len=0,
):
    start = time()
    class_cache = {}
    # initialise counts
    counts = _init_class_counts()
    align_file = os.path.join(workspace, "tmp.bam")

    pair_type_info = {}

    with pysam.AlignmentFile(align_file, "wb", header=header, reference_filename=guide_fa, threads=cpus) as af:
        iter = read_iter(seq_file, default_rgid=default_rgid, cpus=cpus, trim_len=trim_len)
        for seqread_l in iter:
            seqread_r = next(iter, None)
            if seqread_r is None:
                raise ValueError("Collated BAM exhausted between records")

            (read_l, read_r) = (seqread_l.sequence, seqread_r.sequence)
            pair_lookup = f"{read_l}|{read_r}"

            a_l, a_r = None, None

            class_type, hits_l, hits_r, orig_l, orig_r = None, None, None, None, None
            guide_idx = None

            if pair_lookup in class_cache:
                (class_type, guide_idx, hits_l, hits_r, orig_l, orig_r) = class_cache[pair_lookup]
            else:
                (map_l, map_r) = (aligned_results[read_l], aligned_results[read_r])
                # try to order from most likely to least
                (
                    class_type,
                    guide_idx,
                    bt_l,
                    bt_r,
                    orig_l,
                    orig_r,
                ) = classify_read_pair(map_l, map_r, library)

                hits_l = order_hits(map_l[1], bt_l)
                hits_r = order_hits(map_r[1], bt_r)

                class_cache[pair_lookup] = (
                    class_type,
                    guide_idx,
                    hits_l,
                    hits_r,
                    orig_l,
                    orig_r,
                )

            if guide_idx is not None:
                for gidx in guide_idx:
                    library.guides[gidx].count += 1

            if hits_l:
                (a_lst, multi) = to_mapped_reads(seqread_l, ref_ids, library, hits_l, guide_idx=guide_idx, dual=True)
                for a in a_lst:
                    af.write(a)
                if multi:
                    stats.multimap_reads += 1
                stats.mapped_to_guide_reads += 1
            else:
                a = to_alignment(seqread_l, False, [], unmapped=True)
                if class_type == CLASSIFICATION.r_multi_5p:
                    stats.multimap_reads += 1
                else:
                    if orig_l == "multimap":
                        stats.multimap_reads += 1
                    else:
                        stats.unmapped_reads += 1
                af.write(a)
            if hits_r:
                (a_lst, multi) = to_mapped_reads(seqread_r, ref_ids, library, hits_r, guide_idx=guide_idx, dual=True)
                for a in a_lst:
                    af.write(a)
                if multi:
                    stats.multimap_reads += 1
                stats.mapped_to_guide_reads += 1
            else:
                a = to_alignment(seqread_r, False, [], unmapped=True)
                af.write(a)
                if class_type == CLASSIFICATION.r_multi_3p:
                    stats.multimap_reads += 1
                else:
                    if orig_r == "multimap":
                        stats.multimap_reads += 1
                    else:
                        stats.unmapped_reads += 1

            if class_type == CLASSIFICATION.match:
                counts[class_type] += len(guide_idx)
            else:
                counts[class_type] += 1

            if pair_lookup not in pair_type_info:
                classify = {"hit_l": "N", "hit_r": "N", "hit_pair": "N", "count": 0}
                pair_type_info[pair_lookup] = classify
                if class_type == CLASSIFICATION.match:
                    classify["hit_l"] = "Y"
                    classify["hit_r"] = "Y"
                    classify["hit_pair"] = "Y"
                elif class_type in (
                    CLASSIFICATION.aberrant_match,
                    CLASSIFICATION.ambiguous,
                    CLASSIFICATION.f_multi_3p,
                    CLASSIFICATION.r_multi_3p,
                    CLASSIFICATION.f_multi_5p,
                    CLASSIFICATION.r_multi_5p,
                ):
                    classify["hit_l"] = "Y"
                    classify["hit_r"] = "Y"
                elif class_type in (CLASSIFICATION.f_open_3p, CLASSIFICATION.r_open_3p):
                    classify["hit_l"] = "Y"
                elif class_type in (CLASSIFICATION.f_open_5p, CLASSIFICATION.r_open_5p):
                    classify["hit_r"] = "Y"

            pair_type_info[pair_lookup]["count"] += 1

    logging.info(f"Alignment grouping took: {int(time() - start)}s")
    return (counts, align_file, pair_type_info)


def pickles_to_mapset(pickles: List[str], reads: Dict[str, int], aligner: AlignerCpu):
    (aligned_results, multi_map, unique_map, unmap) = ({}, 0, 0, 0)
    for p in pickles:
        logging.info(f"Collating data from {p}")
        batches = ctools.unpickle(p)
        ab: AlignmentBatch
        for ab in batches:
            # unmapped is a simple list
            for s in ab.unmapped:
                aligned_results[s] = ("unmapped", None)
                unmap += reads[s]
            for hits in ab.mapped:
                # function will need to be split out to work via:
                #  library.header.is_single
                best_bt = sg_select_alignment(hits, aligner.rules)
                # for ease of access, common to all hits
                original_seq = hits[0].sm.original_seq
                if len(best_bt) == 0:
                    aligned_results[original_seq] = ("unmapped", None)
                    unmap += reads[original_seq]
                    continue
                elif len(best_bt) > 1:
                    aligned_results[original_seq] = ("multimap", best_bt)
                    multi_map += reads[original_seq]
                    continue
                # don't need to check for existence on this one
                aligned_results[original_seq] = ("unique", best_bt)
                unique_map += reads[original_seq]
    return (aligned_results, multi_map, unique_map, unmap)


def run(
    guidelib,
    queries,
    sample,
    output,
    workspace,
    rules,
    low_count,
    minscore,
    qual_offset,
    cpus,
    reference,
    excludeqcf,
    boundary_mode,
    trimseq,
    chunks,
    loglevel,
):
    (usable_cpu, work_tmp, workspace, boundary_mode) = cli.common_setup(
        loglevel, cpus, workspace, output, boundary_mode
    )
    library = libparser.load(guidelib)
    # returning a list of all guides in a single list
    # this allows us to map reads to both orientations at the same time (set reverse_comp)

    ###
    # input data has to be samtools-collated if hts
    seq_file = readparser.collate(queries, workspace, usable_cpu)
    # read loading needs to return both the unique list of read sequences, and the unique pairs (r1|r2)
    (unique, stats, reads, read_pairs) = readparser.parse_reads(
        seq_file,
        sample=sample,
        cpus=usable_cpu,
        reference=reference,
        exclude_qcfail=excludeqcf,
        paired=True,
        trim_len=trimseq,
    )

    # map the uniq list or individual reads and then use the r1|r2 info to bring the events back together
    aligner = AlignerCpu(
        targets=library.targets,
        rules=rules,
        score_min=minscore,
        rev_comp=True,  # as some reads can be reversed in DG
        match_type=boundary_mode,
    )
    pickles = map_reads(aligner, unique, chunks, cpus, workspace, list(reads.keys()))
    """
    Need to convert alignment batches into a dict by sequence, containing the possible mappings
    """
    (aligned_results, multi_map, unique_map, unmap) = pickles_to_mapset(pickles, reads, aligner)
    logging.info(f"Unique: {unique_map}, Multimap: {multi_map}, Unmapped, {unmap}")
    # * generate the fasta for the guides in workspace
    (guide_fa, header, ref_ids, default_rgid) = guide_header(workspace, library, stats, seq_file)

    (raw_counts, unsorted, pair_type_info) = read_pairs_to_guides(
        workspace,
        aligned_results,
        library,
        seq_file,
        guide_fa,
        header,
        ref_ids,
        default_rgid,
        stats,
        cpus=usable_cpu,
        trim_len=trimseq,
    )
    count_output = f"{output}.counts.tsv"

    # initialise
    if low_count is True:
        stats.low_count_guides_user = {"lt": low_count, "count": 0}
    stats.pair_classifications = raw_counts

    total_hits = 0

    logging.info(f"Writing counts file: {count_output}")
    with open(count_output, "wt") as cout:
        print("##Command: " + stats.command, file=cout)
        print("##Version: " + stats.version, file=cout)
        print("#" + "\t".join(["id", stats.sample_name]), file=cout)

        for g in library.guides:
            print(f"{g.id}\t{g.count}", file=cout)
            if g.count == 0:
                stats.zero_count_guides += 1
            if g.count < 15:
                stats.low_count_guides_lt_15 += 1
            if g.count < 30:
                stats.low_count_guides_lt_30 += 1
            if low_count is True and g.count < low_count:
                stats.low_count_guides_user["count"] += 1
            total_hits += g.count

    stats.total_guides = len(library.guides)
    stats.mean_count_per_guide = round(total_hits / len(library.guides), 2)

    stats_output = f"{output}.stats.json"
    logging.info(f"Writing statistics file: {stats_output}")
    with open(stats_output, "wt") as jout:
        print(json.dumps(stats.__dict__, sort_keys=True, indent=2), file=jout)

    seqclass_output = f"{output}.query_class.tsv.gz"
    logging.info(f"Writing query sequence classifications: {seqclass_output}")
    with gzip.open(seqclass_output, "wt") as scot:
        print(f"#read_seqs\thit_l\thit_r\thit_pair\tcount", file=scot)
        for k in sorted(pair_type_info.keys()):
            print(
                f'{k}\t{pair_type_info[k]["hit_l"]}\t{pair_type_info[k]["hit_r"]}\t{pair_type_info[k]["hit_pair"]}\t{pair_type_info[k]["count"]}',
                file=scot,
            )

    hts_sort_n_index(unsorted, guide_fa, output, workspace, cpus=usable_cpu)

    if not work_tmp:
        shutil.rmtree(workspace)



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import gzip
import json
import logging
import os
import sys
from typing import Dict
from typing import Tuple

from pycroquet.classes import Guide
from pycroquet.classes import Library
from pycroquet.classes import Stats

# these are all in the root of guide object
COLS_REQ = ["id", "sgrna_ids", "sgrna_seqs", "gene_pair_id"]


def _fmt_single(guide: Guide, count: int) -> str:
    to_join = []
    for c in COLS_REQ:
        attr = getattr(guide, c)
        if type(attr) is list:
            attr = "|".join(attr)
        to_join.append(attr)

    to_join.append(str(int(guide.unique)))  # better for the output
    to_join.append(str(count))
    return "\t".join(to_join)


def _header(sample: str) -> str:
    header = "#"
    for c in COLS_REQ:
        header += f"{c}\t"
    header += "unique_guide\t"
    header += f"reads_{sample}"
    return header


def guide_counts_single(
    library: Library,
    guide_results: Dict[str, int],
    output: str,
    stats: Stats,
    low_count: int = None,
) -> Tuple[str, int]:
    """
    Generates the primary result file, reads hitting guides
    """
    if low_count is True:
        stats.low_count_guides_user = {"lt": low_count, "count": 0}
    count_output = f"{output}.counts.tsv"
    logging.info(f"Writing counts file: {count_output}")
    count_total = 0
    # target_to_guides = library.target_to_guides  # can't remember why I included this
    with open(count_output, "wt") as cout:
        print("##Command: " + stats.command, file=cout)
        print("##Version: " + stats.version, file=cout)
        print(_header(stats.sample_name), file=cout)
        # guide_results = {seq : count,...}
        for guide in library.guides:
            # need to output columns in a stable order, core fields then alpha?
            guide_seq = guide.sgrna_seqs[0]
            count = guide_results.get(guide_seq, 0)
            if count < 30:
                stats.low_count_guides_lt_30 += 1
                if count < 15:
                    stats.low_count_guides_lt_15 += 1
                if count == 0:
                    stats.zero_count_guides += 1
            if low_count is True and count < low_count:
                stats.low_count_guides_user["count"] += 1
            print(_fmt_single(guide, count), file=cout)
            count_total += count

    stats.mean_count_per_guide = round(count_total / len(library.guides), 2)

    stats_output = f"{output}.stats.json"
    logging.info(f"Writing statistics file: {stats_output}")
    with open(stats_output, "wt") as jout:
        print(json.dumps(stats.__dict__, sort_keys=True, indent=2), file=jout)
    return (count_output, count_total)


def query_counts(
    query_dict: Dict[str, int],
    stats: Stats,
    output: str,
):
    """
    This generates a file with the number of incidents of the same query sequence
    """
    count_output = f"{output}.query_counts.tsv.gz"
    logging.info(f"Writing query counts file: {count_output}")
    with gzip.open(count_output, "wt") as cout:
        print("##Command: " + stats.command, file=cout)
        print("##Version: " + stats.version, file=cout)
        print("#QUERY\tCOUNT", file=cout)
        for k in sorted(query_dict.keys()):
            print(f"{k}\t{query_dict[k]}", file=cout)



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import copy
import gzip
import hashlib
import logging
import os
from array import array
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple

import pysam
from pygas.classes import Backtrack
from pygas.matrix import revcomp

from pycroquet.classes import Library
from pycroquet.classes import Seqread
from pycroquet.classes import Stats
from pycroquet.constants import EXT_TO_HTS
from pycroquet.htscomm import hts_reader
from pycroquet.htscomm import hts_sort_n_index
from pycroquet.readparser import is_gzip
from pycroquet.readparser import parse_fq_header


def _hts_iter(
    seq_file,
    mode,
    exclude_qcfail,
    reverse,
    reference,
    cpus=1,
    default_rgid=None,
    trim_len=0,
) -> Iterator[Seqread]:
    sam = hts_reader(seq_file, mode, cpus, reference)
    for read in sam.fetch(until_eof=True):
        if read.is_secondary or read.is_supplementary:
            continue
        if read.is_qcfail and exclude_qcfail:
            continue
        seq = read.get_forward_sequence()
        qual = read.get_forward_qualities()
        if trim_len:
            seq = seq[0:trim_len]
            qual = qual[0:trim_len]
        if reverse:
            seq = revcomp(seq)
            qual = qual[::-1]
        member = (1 if read.is_read1 else 2) if read.is_paired else None
        rgid = read.get_tag("RG") if read.has_tag("RG") else default_rgid
        yield Seqread(
            qname=read.query_name,
            sequence=seq,
            member=member,
            qual=qual,
            qc_fail=read.is_qcfail,
            rgid=rgid,
        )
    sam.close()


def _fq_iter(ifh, exclude_qcfail, reverse, offset_override=None, default_rgid=None, trim_len=0) -> Iterator[Seqread]:
    header = ifh.readline()
    while header:
        seq = ifh.readline().strip()
        _ = ifh.readline()  # throw away separator
        qual_str = ifh.readline().strip()
        if trim_len:
            seq = seq[0:trim_len]
            qual_str = qual_str[0:trim_len]
        (qname, member, qc_fail, phred_offset) = parse_fq_header(header.strip())  # for validation only here
        if qc_fail and exclude_qcfail:
            header = ifh.readline()
            continue
        if offset_override:
            phred_offset = offset_override
        phred_arr = [i - phred_offset for i in map(ord, qual_str)]

        if reverse:
            phred_arr.reverse()
            seq = revcomp(seq)
        yield Seqread(
            qname=qname,
            member=member,
            sequence=seq,
            qual=array("i", phred_arr),
            qc_fail=qc_fail,
            rgid=default_rgid,
        )
        header = ifh.readline()
    ifh.close()


def read_iter(
    seq_file: str,
    reverse: bool = False,
    exclude_qcfail: bool = False,
    reference: str = None,
    offset=None,
    default_rgid=None,
    cpus=1,
    trim_len=0,
) -> Iterator[Seqread]:
    base_iter = None
    ext = os.path.splitext(seq_file)[1]
    if ext in EXT_TO_HTS:
        base_iter = _hts_iter(
            seq_file,
            EXT_TO_HTS[ext],
            exclude_qcfail,
            reverse,
            reference,
            default_rgid=default_rgid,
            cpus=cpus,
            trim_len=trim_len,
        )
    else:
        fq_fo = None
        if is_gzip(seq_file):
            fq_fo = gzip.open(seq_file, "rt")
        else:
            fq_fo = open(seq_file, "rt")
        base_iter = _fq_iter(
            fq_fo,
            exclude_qcfail,
            reverse,
            offset_override=offset,
            default_rgid=default_rgid,
            trim_len=trim_len,
        )
    return base_iter


def to_alignment(seqread: Seqread, reverse: bool, tags: List[Tuple[str, str]], unmapped=False) -> pysam.AlignedSegment:
    tags.append(("RG", seqread.rgid))
    a = pysam.AlignedSegment()
    if reverse:
        a.query_sequence = revcomp(seqread.sequence)
        seqread.qual.reverse()
        a.query_qualities = seqread.qual
        a.is_reverse = True
    else:
        a.query_sequence = seqread.sequence
        a.query_qualities = seqread.qual

    a.query_name = seqread.qname
    a.tags = tags
    a.is_qcfail = seqread.qc_fail
    a.is_unmapped = unmapped

    if seqread.member:
        a.is_paired = True
        if seqread.member == 1:
            a.is_read1 = True
            a.is_read2 = False
        elif seqread.member == 2:
            a.is_read1 = False
            a.is_read2 = True
        else:
            raise ValueError("Only, unpaired, or paired 1/2 reads are anticipated")
    else:
        a.is_paired = False
    return a


def rg_pg(seq_file: str, stats: Stats) -> Tuple[List[Dict[str, str]], List[Dict[str, str]]]:
    ext = os.path.splitext(seq_file)[1]
    rg = {"ID": "1", "SM": stats.sample_name}
    prog = " ".join(stats.command.split()[0:2])
    pg = {
        "ID": "pycroquet",
        "PN": prog,
        "CL": stats.command,
        "VN": stats.version,
        "DS": "python Crispr Read to Oligo QUantification Enhancement Tool",
    }

    if ext not in EXT_TO_HTS:
        return ([rg], [pg])

    sam = hts_reader(seq_file, EXT_TO_HTS[ext], 1)
    header = sam.header.as_dict()

    pgs = []
    if "PG" in header:
        pg["PP"] = header["PG"][-1]["ID"]
        pgs = header["PG"]
    pgs.append(pg)

    rgs = []
    if "RG" not in header:
        rgs.append(rg)
    else:
        for r in header["RG"]:
            r["SM"] = stats.sample_name
            rgs.append(r)

    return (rgs, pgs)


def guide_fasta(library: Library, guide_fa: str, index: bool = False):
    sq_data = []
    ref_ids = {}
    with open(guide_fa, "wt") as gfa:
        ref_idx = 0
        for g in library.guides:
            for sg_idx, sgrna_id in enumerate(g.sgrna_ids):
                if sgrna_id in ref_ids:
                    continue
                print(f">{sgrna_id}", file=gfa)
                print(f"{g.sgrna_seqs[sg_idx]}", file=gfa)
                sq_data.append(
                    {
                        "LN": len(g.sgrna_seqs[sg_idx]),
                        "SN": sgrna_id,
                        "M5": hashlib.md5(g.sgrna_seqs[sg_idx].encode("utf-8")).hexdigest(),
                    }
                )
                ref_ids[sgrna_id] = ref_idx
                ref_idx += 1
    pysam.faidx(guide_fa)
    return (sq_data, ref_ids)


def guide_header(workspace: str, library: Library, stats: Stats, seq_file: str) -> Tuple[str, dict, Dict[str, int]]:
    guide_fa = os.path.join(workspace, "guides.fa")
    (sq_data, ref_ids) = guide_fasta(library, guide_fa)

    (rg, pg) = rg_pg(seq_file, stats)
    default_rgid = None
    if len(rg) == 1:
        default_rgid = rg[0]["ID"]
    header = {"HD": {"VN": "1.0"}, "SQ": sq_data, "RG": rg, "PG": pg}
    return (guide_fa, header, ref_ids, default_rgid)


def sm_from_guide_hit(library: Library, hit: Backtrack, strand: str, mapq: int = 60) -> Tuple[Dict[str, str], int]:
    sgrna_ids = library.sgrna_ids_by_seq(hit.sm.target)
    sa_set = {}
    for sgrna_id in sgrna_ids:
        sa_set[sgrna_id] = f"{sgrna_id},{hit.t_pos},{strand},{hit.cigar},{mapq},{hit.nm}"
    return sa_set


def to_mapped_reads(
    seqread,
    ref_ids,
    library: Library,
    hits: List[Backtrack],
    reverse_in=None,
    strand_in=None,
    guide_idx=None,
    dual=False,
) -> Tuple[List[pysam.AlignedSegment], bool]:
    mapq = 60
    if len(hits) > 1:
        mapq = 0
    sa_set = {}
    strand = strand_in
    reverse = reverse_in
    for hit in hits:
        if strand_in is None:
            strand = "-" if hit.sm.reversed else "+"
        this_sa_set = sm_from_guide_hit(library, hit, strand, mapq)
        sa_set = sa_set | this_sa_set  # merge 2 dicts requires >=3.9
    is_secondary = False
    alignments = []
    sa_set_len = len(sa_set)
    for hit in hits:
        ref_names = library.sgrna_ids_by_seq(hit.sm.target)
        tags = [("NM", hit.nm), ("MD", hit.md), ("AS", hit.sm.score)]
        if guide_idx is not None and is_secondary is False:
            # only apply to the primary mapping
            tags.append(("YG", ";".join([library.guides[i].id for i in guide_idx])))
        for ref_name in ref_names:
            local_tags = copy.copy(tags)
            if sa_set_len > 1:
                local_tags.append(("NH", sa_set_len))
                if sa_set_len <= 10:
                    local_tags.append(
                        (
                            "SA",
                            ";".join([v for k, v in sa_set.items() if k != ref_name]),
                        )
                    )

            if reverse_in is None:
                reverse = hit.sm.reversed

            a = to_alignment(seqread, reverse, local_tags)
            # complete mapped info
            a.is_secondary = is_secondary
            a.reference_id = ref_ids[ref_name]
            a.reference_start = hit.t_pos - 1  # zero based
            a.cigarstring = hit.cigar
            a.mapping_quality = mapq
            alignments.append(a)
            if dual:
                break
            is_secondary = True
        if dual:
            break
    return (alignments, sa_set_len > 1)


def write_alignments(
    align_file,
    header,
    guide_fa,
    cpus,
    seq_file,
    reverse,
    exclude_qcfail,
    reference,
    qual_offset,
    aligned_results,
    library,
    ref_ids,
    strand,
    default_rgid,
    skipped=None,
):
    """
    own function to allow try due to incorrect qual threshold use
    """
    with pysam.AlignmentFile(align_file, "wb", header=header, reference_filename=guide_fa, threads=cpus) as af:
        # get the common-iterator
        iter = read_iter(
            seq_file,
            reverse,
            exclude_qcfail,
            reference,
            offset=qual_offset,
            default_rgid=default_rgid,
        )
        skipped_reads = 0
        for seqread in iter:
            # * Pair the sequence to the mapped records
            if seqread.sequence not in aligned_results:
                if skipped is not None:
                    skipped_reads += 1
                    if skipped_reads > skipped:
                        raise ValueError(
                            "Source read not found in aligned data and number of reads skipped due to length has beed exceeded. Has input been modified during execution?"
                        )
                    a = to_alignment(seqread, False, [], unmapped=True)
                    af.write(a)
                    continue

                raise ValueError("Source read not found in aligned data, has input been modified during execution?")
            (hit_type, hits) = aligned_results[seqread.sequence]

            # add the basics to the alignment

            a = None
            if hit_type in ("unique", "multimap"):
                (a_lst, _) = to_mapped_reads(seqread, ref_ids, library, hits, reverse, strand_in=strand)
                for a in a_lst:
                    af.write(a)
            elif hit_type == "unmapped":
                a = to_alignment(seqread, reverse, [], unmapped=True)
                af.write(a)


def reads_to_hts(
    library: Library,
    aligned_results: Dict[str, Tuple[str, List[Backtrack]]],
    seq_file: str,
    qual_offset: str,
    workspace: str,
    stats: Stats,
    output: str,
    cpus: int,
    reverse: bool = False,
    exclude_qcfail: bool = False,
    reference: str = None,
):
    if qual_offset:
        qual_offset = int(qual_offset)
    strand = "-" if reverse else "+"
    # * generate the fasta for the guides in workspace
    (guide_fa, header, ref_ids, default_rgid) = guide_header(workspace, library, stats, seq_file)

    # * Create an AlignmentFile with relevant "sequences"
    align_file = os.path.join(workspace, "temp.bam")

    logging.info(f"Writing alignment file: {align_file} (intermediate)")

    sam_threads = cpus if cpus < 4 else 4

    try:
        write_alignments(
            align_file,
            header,
            guide_fa,
            sam_threads,
            seq_file,
            reverse,
            exclude_qcfail,
            reference,
            qual_offset,
            aligned_results,
            library,
            ref_ids,
            strand,
            default_rgid,
            skipped=stats.length_excluded_reads,
        )
    except OverflowError as e:
        seq_extn = os.path.splitext(seq_file)[1].lower
        if str(e) != "unsigned byte integer is less than minimum" and seq_extn in (
            ".sam",
            ".bam",
            ".cram",
        ):
            raise e
        logging.error("FASTQ: phred offset not deduced correctly from readname, attempting to handle internally")
        write_alignments(
            align_file,
            header,
            guide_fa,
            sam_threads,
            seq_file,
            reverse,
            exclude_qcfail,
            reference,
            33,
            aligned_results,
            library,
            ref_ids,
            strand,
            default_rgid,
            skipped=stats.length_excluded_reads,
        )

    hts_sort_n_index(align_file, guide_fa, output, workspace, cpus=sam_threads)



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import os
import sys
from array import array
from dataclasses import dataclass
from dataclasses import field
from typing import Dict
from typing import Final
from typing import List

import pkg_resources


@dataclass
class Classification:
    match: Final = "match"  # same vector F/R
    aberrant_match: Final = "aberrant_match"  # same vector pair, aberrant orientation
    f_multi_3p: Final = "f_multi_5p"  # 5p mapped F, 3p multihit
    f_multi_5p: Final = "f_multi_5p"  # 3p mapped F, 5p multihit
    r_multi_3p: Final = "r_multi_5p"  # 5p mapped R, 3p multihit
    r_multi_5p: Final = "r_multi_5p"  # 3p mapped R, 5p multihit
    f_open_3p: Final = "f_open_3p"  # 5p mapped F, 3p open (unmapped)
    f_open_5p: Final = "f_open_5p"  # 3p mapped F, 5p open (unmapped)
    r_open_3p: Final = "f_open_3p"  # 5p mapped R, 3p open (unmapped)
    r_open_5p: Final = "f_open_5p"  # 3p mapped R, 5p open (unmapped)
    swap: Final = "swap"  # multi vector, uniq mapped
    ambiguous: Final = "ambiguous"  # both ends multi hit
    no_match: Final = "no_match"  # multi/unmapped either end


"""
Feature count total_number
MATCH: 18379264 20617796  // The match is perfect R1 F(Forward) and R2 R(reverse)
MATCHFF: 0 20617796.      // Match but R1 is F and R2 is F
MATCHRF: 0 20617796       // Match but R1 is R and R2 is F
MATCHRR: 0 20617796       // Match but R1 is R and R2 is R
SWAPFF: 0 20617796        // R1 is mapping a vector F and R2 is mapping another vector F
SWAPFR: 404446 20617796   // R1 is mapping a vector F and R2 is mapping another vector R
SWAPRF: 0 20617796        // R1 is mapping a vector R and R2 is mapping another vector F
SWAPRR: 0 20617796        // R1 is mapping a vector R and R2 is mapping another vector R
OEM1F: 1063485 20617796  //  R1 is mapping to a vector F and R2 is not mapping
OEM1R: 0 20617796        //  R1 is mapping to a vector R and R2 is not mapping
OEM2F: 0 20617796        //  R2 is mapping to a vector F and R1 is not mapping
OEM2R: 722427 20617796   //  R2 is mapping to a vector R and R1 is not mapping
NOMATCH: 48174 20617796 // Both R1 and R2 are not mapping
"""


@dataclass
class LibraryHeader:
    column_map: dict
    column_list: List[str]
    column_separators: dict
    required_cols: List[str]
    config: dict
    info_items: dict = None
    is_single: bool = False
    reverse_read_order: bool = False


@dataclass
class Guide:
    idx: int
    id: str = None
    sgrna_ids: List[str] = None
    sgrna_seqs: List[str] = None
    gene_pair_id: str = None
    sgrna_strands: List[int] = None
    other: dict[str, str] = field(default_factory=dict)
    unique: bool = True
    count: int = 0
    _composite_sgrna_seq = None

    def composite_sgrna_seq(self) -> str:
        if self._composite_sgrna_seq is None:
            self._composite_sgrna_seq = "|".join(self.sgrna_seqs)
        return self._composite_sgrna_seq


@dataclass
class Library:
    """
    header: the header object
    guides: guide detail objects
    targets: unique guide sequences
    target_to_guides: mappings of target sequences back to guides
    """

    header: LibraryHeader
    guides: List[Guide]
    targets: List[str]
    target_to_guides: Dict[str, List[int]]
    _sgrna_ids_by_seq: Dict[str, int] = None
    _guide_by_sgrna_set: Dict[str, int] = None

    def min_target_len(self) -> int:
        return len(min(self.targets, key=len))

    def sgrna_ids_by_seq(self, target_seq):
        """
        get the target by the 0,1... position in a guide
        """
        if self._sgrna_ids_by_seq is None:
            sgrna_ids_by_seq = {}
            frags = len(self.guides[0].sgrna_ids)
            for g in self.guides:
                for i in range(0, frags):
                    g_seq = g.sgrna_seqs[i]
                    g_id = g.sgrna_ids[i]
                    if g_seq not in sgrna_ids_by_seq:
                        sgrna_ids_by_seq[g_seq] = {}
                    sgrna_ids_by_seq[g_seq][g_id] = None
            # reduce to list of keys, note data type changes from dict to list
            for g_seq in sgrna_ids_by_seq:
                sgrna_ids_by_seq[g_seq] = sorted(sgrna_ids_by_seq[g_seq].keys())
            self._sgrna_ids_by_seq = sgrna_ids_by_seq
        return self._sgrna_ids_by_seq[target_seq]

    def guide_by_sgrna_set(self, seq_l, seq_r) -> List[int]:
        match = f"{seq_l}|{seq_r}"
        if self._guide_by_sgrna_set is None:
            data = {}
            for i, g in enumerate(self.guides):
                composite_guide = g.composite_sgrna_seq()
                if composite_guide not in data:
                    data[composite_guide] = []
                data[composite_guide].append(i)
            self._guide_by_sgrna_set = data
        return self._guide_by_sgrna_set.get(match)


@dataclass
class Stats:
    """
    Read and guide statistics
    """

    total_reads: int = 0
    total_pairs: int = 0
    vendor_failed_reads: int = 0
    mapped_to_guide_reads: int = 0
    # reads that map equally well to multiple guides, mainly for SG data
    multimap_reads: int = 0
    unmapped_reads: int = 0
    total_guides: int = 0
    zero_count_guides: int = 0
    low_count_guides_lt_15: int = 0
    low_count_guides_lt_30: int = 0
    mean_count_per_guide: int = 0
    length_excluded_reads: int = None
    low_count_guides_user: Dict[str, int] = None
    reversed_reads: bool = False
    version: str = None
    command: str = None
    sample_name: str = None
    pair_classifications: Dict[str, int] = None

    def __post_init__(self):
        self.command = ""
        for i, e in enumerate(sys.argv):
            if i == 0:
                self.command += f"{os.path.basename(e)}"
                continue
            self.command += f" {e}"
        self.version = pkg_resources.require(__name__.split(".")[0])[0].version


@dataclass
class Seqread:
    """
    Raw sequence info required for mapping outputs
    """

    qname: str
    sequence: str
    qual: array
    rgid: str
    qc_fail: bool = False
    member: int = None



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
"""
For shared constants to prevent circular import issues
"""
from typing import Final

EXT_TO_HTS: Final = {
    ".bam": "rb",
    ".cram": "rc",
    ".sam": "r",
}



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import logging
import os
import sys
import tempfile
from functools import wraps

import click
import pkg_resources  # part of setuptools
from click_option_group import OptionGroup

from pycroquet import dualguide
from pycroquet import libparser
from pycroquet import main
from pycroquet import readwriter
from pycroquet import sge as pysge
from pycroquet import singleguide
from pycroquet import tools as ctools

LOG_LEVELS = ("WARNING", "INFO", "DEBUG")

HELP_TRIMSEQ = "Trim reads back to use first N bases only. This is a destructive process, the alignment CRAM will not include the full read sequence."
HELP_GUIDELIB = "Expanded guide library definition tsv file with optional headers (common format for single/dual/other)"
HELP_QUERIES = "Query sequence file (fastq[.gz], sam, bam, cram)"
HELP_SAMPLE = (
    "Sample name to apply to count column, required for fastq, interrogate header for others when not defined."
)
HELP_OUTPUT = "Final output to this filename prefix"
HELP_WORKSPACE = "Workspace for temporary files"
HELP_RULES = "Rules for allowing fuzzy matching (no rules = fast exact/substr match only), e.g. MM=2 mismatched bases, MDI allows 1 base of mismatch, deletion and insertion."
HELP_LOW_COUNT = (
    "*.stats.json includes low_count_guides_lt_{15,30}, this option allow specification of an additional cut-off."
)
HELP_QUAL_OFFSET = "Specify phread offset (for fastq) if detection by readname fails"
HELP_CPUS = "CPUs to use (0 to detect)"
HELP_SGE_UNIQUE = "Only generate the unique sequence counts file, then exit (--guide can be omitted)"
HELP_CHUNKS = "Reads per mapping block"
HELP_MINSCORE = "Minimum score to retain, regardless of rule penalties.  Perfect match has score equal to query length."
HELP_REFERENCE = "Required for cram"
HELP_EXCLUDEQCF = "Exclude reads that fail QC (sam/bam/cram, CASAVA fastq only)"
HELP_NO_ALIGNMENT = "Do not output cram alignments"
HELP_BOUNDARY = "Control boundary matching types, see end of options"
HELP_FASTA = "Write fasta to this file"

HELP_EPILOG = """
Additional option info:

-b/--boundary-mode

\b
  all   = all combinations
  exact = requires all guides and queries to be same length
  TinQ  = target in query
  QinT  = query in target
"""


def _log_setup(loglevel):  # pragma: no cover
    sh_info = logging.StreamHandler(stream=sys.stdout)
    sh_info.setLevel(logging.DEBUG)
    sh_error = logging.StreamHandler(stream=sys.stderr)
    sh_error.setLevel(logging.ERROR)

    logging.basicConfig(
        level=getattr(logging, loglevel.upper()), format="%(levelname)s: %(message)s", handlers=[sh_info, sh_error]
    )


def common_setup(loglevel, cpus, workspace, output, boundary_mode):
    _log_setup(loglevel)
    usable_cpu = cpus if cpus > 0 else len(os.sched_getaffinity(0))
    work_tmp = None
    if workspace:
        os.makedirs(workspace, exist_ok=True)
    else:
        work_tmp = tempfile.TemporaryDirectory()
        workspace = work_tmp.name
    outfolder = os.path.dirname(os.path.abspath(output))
    if not os.path.exists(outfolder):
        os.makedirs(outfolder, exist_ok=True)
    boundary_mode = ctools.boundary_mode(boundary_mode)
    return (usable_cpu, work_tmp, workspace, boundary_mode)


def _file_exists():
    return click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    )


def chunk_default(f):
    @click.option(
        "--chunks", required=False, type=int, default=main.READ_CHUNK_INT, show_default=True, help=HELP_CHUNKS
    )
    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


def sge_extra(f):
    @click.option("--unique", "unique_only", required=False, type=bool, is_flag=True, help=HELP_SGE_UNIQUE)
    @click.option(
        "--chunks", required=False, type=int, default=main.READ_CHUNK_SGE_INT, show_default=True, help=HELP_CHUNKS
    )
    @click.option(
        "-n",
        "--no-alignment",
        required=False,
        default=False,
        type=bool,
        help=HELP_NO_ALIGNMENT,
        show_default=True,
        is_flag=True,
    )
    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


def common_params(f):
    @click.option("-g", "--guidelib", required=False, default=None, type=_file_exists(), help=HELP_GUIDELIB)
    @click.option("-q", "--queries", required=True, type=_file_exists(), help=HELP_QUERIES)
    @click.option("-s", "--sample", required=False, type=str, help=HELP_SAMPLE)
    @click.option(
        "-o",
        "--output",
        required=True,
        type=click.Path(exists=False, file_okay=True, resolve_path=True),
        help=HELP_OUTPUT,
    )
    @click.option(
        "-w",
        "--workspace",
        required=False,
        type=click.Path(exists=False, dir_okay=True, resolve_path=True),
        help=HELP_WORKSPACE,
    )
    @click.option(
        "--rules",
        required=False,
        type=str,
        default=[],
        multiple=True,
        help=HELP_RULES,
        show_default=True,
    )
    @click.option("--low_count", required=False, type=int, default=None, help=HELP_LOW_COUNT, show_default=True)
    @click.option(
        "--qual_offset",
        required=False,
        type=click.Choice(["33", "64"]),
        default=None,
        help=HELP_QUAL_OFFSET,
        show_default=True,
    )
    @click.option("-c", "--cpus", required=False, type=int, default=1, show_default=True, help=HELP_CPUS)
    @click.option("-m", "--minscore", required=False, type=int, default=15, help=HELP_MINSCORE, show_default=True)
    @click.option("-r", "--reference", required=False, type=_file_exists(), help=HELP_REFERENCE)
    @click.option(
        "-e",
        "--excludeqcf",
        required=False,
        default=False,
        type=bool,
        help=HELP_EXCLUDEQCF,
        show_default=True,
        is_flag=True,
    )
    @click.option(
        "-b",
        "--boundary-mode",
        required=False,
        default="all",
        help=HELP_BOUNDARY,
        show_default=True,
        type=click.Choice(("all", "exact", "TinQ", "QinT"), case_sensitive=False),
    )
    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


optgroup_debug = OptionGroup("\nDebug options", help="Options specific to troubleshooting, testing and debugging")


def debug_params(f):
    @optgroup_debug.option(
        "-l",
        "--loglevel",
        required=False,
        default="INFO",
        show_default=True,
        type=click.Choice(LOG_LEVELS, case_sensitive=False),
        help="Set logging verbosity",
    )
    @wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


@click.group()
@click.version_option(pkg_resources.require(__name__.split(".")[0])[0].version)
def cli():  # pragma: no cover
    pass


@cli.command(epilog=HELP_EPILOG)
@common_params
@chunk_default
@click.option(
    "-n",
    "--no-alignment",
    required=False,
    default=False,
    type=bool,
    help=HELP_NO_ALIGNMENT,
    show_default=True,
    is_flag=True,
)
@debug_params
def single_guide(*args, **kwargs):
    """
    Single guide - map read file to library guides and output counts, statistics and alignments files.
    """
    singleguide.run(*args, **kwargs)


@cli.command(epilog=HELP_EPILOG)
@common_params
@chunk_default
@click.option(
    "-t",
    "--trimseq",
    required=False,
    type=int,
    default=0,
    show_default=True,
    help=HELP_TRIMSEQ,
)
@debug_params
def dual_guide(*args, **kwargs):
    """
    Dual guide - map read file to library guides and output counts, statistics and alignments files.
    """
    dualguide.run(*args, **kwargs)


@cli.command(epilog=HELP_EPILOG)
@common_params
@sge_extra
@debug_params
def long_read(*args, **kwargs):
    """
    log-read - map read file to library guides and output counts, statistics and alignments files.
    Different to single-guide due to read filtering and other defaults
    """
    pysge.run(*args, **kwargs)


@cli.command()
@click.option("-g", "--guidelib", required=True, type=_file_exists(), help=HELP_GUIDELIB)
@click.option(
    "-f",
    "--fasta",
    required=True,
    type=click.Path(exists=False, file_okay=True, resolve_path=True),
    help=HELP_FASTA,
)
@debug_params
def guides_to_fa(guidelib, fasta, loglevel):
    """
    Convert guide library to fasta file, mainly for debug use via "samtools tview".
    """
    _log_setup(loglevel)
    readwriter.guide_fasta(libparser.load(guidelib), fasta, index=True)



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
"""
For common hts file actions
"""
import logging
import os
from tempfile import TemporaryDirectory

import pysam


def hts_reader(seq_file, mode, cpus, reference=None) -> pysam.AlignmentFile:
    hts_cpus = cpus if cpus < 4 else 4
    save = pysam.set_verbosity(0)
    sam = pysam.AlignmentFile(
        seq_file,
        mode=mode,
        reference_filename=reference,
        require_index=False,
        threads=hts_cpus,
        check_sq=False,
    )
    pysam.set_verbosity(save)
    return sam


def hts_sort_n_index(unsorted, guide_fa, output_stub, workspace, cpus=1):
    sam_threads = 4
    if cpus < sam_threads:
        sam_threads = cpus
    aligned_cram = f"{output_stub}.cram"
    logging.info(f"Sorting alignments to: {aligned_cram}")
    with TemporaryDirectory(prefix="sort_to_cram", dir=workspace) as td:
        pysam.sort(
            "--no-PG",
            "-@",
            str(sam_threads),
            "--reference",
            guide_fa,
            "--output-fmt",
            "CRAM,no_ref=1",
            "-T",
            os.path.join(td, "samsort"),
            "-o",
            aligned_cram,
            unsorted,
        )
    logging.info(f"Generating alignment index: {aligned_cram}.crai")
    pysam.index(aligned_cram)



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import logging
import multiprocessing as mp
import random
from functools import partial
from time import time
from typing import Dict
from typing import List
from typing import Tuple

from pygas.alignercpu import AlignerCpu
from pygas.classes import AlignmentBatch
from pygas.classes import Backtrack

import pycroquet.tools as ctools
from pycroquet import readparser
from pycroquet.classes import Library
from pycroquet.classes import Stats
from pycroquet.countwriter import guide_counts_single

READ_CHUNK_INT = 20000
READ_CHUNK_SGE_INT = 1000


def map_thread(query_seqs: List[str], aligner: AlignerCpu) -> AlignmentBatch:
    return aligner.align_queries(query_seqs, keep_matrix=False)


def mapping_by_chunk(aligner: AlignerCpu, query_seqs, read_chunk, workspace, usable_cpu) -> List[str]:
    # randomizes read order to distribute harder tasks, result is still reproducible
    random.Random().shuffle(query_seqs)
    # library.targets  # the targets for pygas
    # query_seqs  # the queries for pygas
    seq_sets = ctools.chunks(query_seqs, read_chunk * usable_cpu)
    del query_seqs  # make sure that it is removed as it's been duplicated

    # setup for multi-process

    if __name__ == "__main__" and mp.get_start_method(allow_none=True) is None:
        mp.set_start_method("spawn")

    start = time()
    was = start
    pickled_files = []
    for seq_set in seq_sets:
        this_seq_set = list(ctools.chunks(seq_set, read_chunk))
        results = None
        if usable_cpu == 1 or len(this_seq_set) == 1:  # save overhead
            results = [map_thread(this_seq_set[0], aligner)]
        else:
            with mp.Pool(processes=usable_cpu) as pool:
                results = pool.map(partial(map_thread, aligner=aligner), this_seq_set)
        now = time()
        logging.info(f"{usable_cpu} CPUs processed {len(seq_set)} reads in {int(now - was)}s (wall)")
        was = now

        pickled_files.append(ctools.pickle_this(workspace, "pre_matrix_{:05d}".format(len(pickled_files) + 1), results))
        del results
        # logging.warning("exiting loop early for development purposes")
        # break
    return pickled_files


def map_reads(
    aligner: AlignerCpu,
    unique: int,
    read_chunk: int,
    cpus: int,
    workspace: str,
    query_seqs: List[str],
):
    if unique < read_chunk * cpus:
        new_chunk = int(unique / cpus) + 1
        logging.warning(f"--chunks value {read_chunk} rescaled to {new_chunk} to utilise all CPUs")
        read_chunk = new_chunk

    pickles = mapping_by_chunk(aligner, query_seqs, read_chunk, workspace, cpus)
    return pickles


def sg_select_alignment(hits: List[Backtrack], rules: List[str]) -> List[Backtrack]:
    """
    Single guide is very straight forward for selecting the mapping
    - best score that fulfils the rules
    """
    best = []
    best_score = 0
    for bt in hits:
        sm = bt.sm
        if sm.score > best_score:
            best = []
            best_score = sm.score
        elif sm.score < best_score:
            continue
        if bt.pass_rules(rules):
            best.append(bt)
    return best


def process_reads(
    library: Library,
    seqfile: str,
    workspace,
    rules: List[str],
    minscore: int,
    cpus: int,
    read_chunk: int,
    sample=None,
    reference=None,
    exclude_qcfail=None,
    reverse=False,
    exclude_by_len=None,
    boundary_mode=3,
) -> Tuple[Dict[str, int], Dict[str, Tuple[str, List[Backtrack]]], Stats]:
    (unique, stats, query_dict, _) = readparser.parse_reads(
        seqfile,
        sample=sample,
        cpus=cpus,
        reference=reference,
        exclude_qcfail=exclude_qcfail,
        exclude_by_len=exclude_by_len,
    )
    aligner = AlignerCpu(
        targets=library.targets,
        rules=rules,
        score_min=minscore,
        rev_comp=reverse,
        match_type=boundary_mode,
    )

    pickles = map_reads(aligner, unique, read_chunk, cpus, workspace, list(query_dict.keys()))

    # here we are collecting the results into a dict so we can assess them as we pass over the read file again
    aligned_results = {}
    guide_results = {}
    (mapped, multimap, unmapped) = (0, 0, 0)
    for p in pickles:
        logging.info(f"Collating data from {p}")
        batches = ctools.unpickle(p)
        ab: AlignmentBatch
        for ab in batches:
            # unmapped is a simple list
            for s in ab.unmapped:
                aligned_results[s] = ("unmapped", None)
                unmapped += query_dict[s]
            for hits in ab.mapped:
                # function will need to be split out to work via:
                #  library.header.is_single
                best_bt = sg_select_alignment(hits, aligner.rules)
                # for ease of access, common to all hits
                original_seq = hits[0].sm.original_seq
                if len(best_bt) == 0:
                    aligned_results[original_seq] = ("unmapped", None)
                    unmapped += query_dict[original_seq]
                    continue
                elif len(best_bt) > 1:
                    aligned_results[original_seq] = ("multimap", best_bt)
                    multimap += query_dict[original_seq]
                    continue

                # don't need to check for existence on this one
                aligned_results[original_seq] = ("unique", best_bt)
                best_align = best_bt[0]
                # but do here as read with mm/d/i can map to the guide
                if best_align.sm.target not in guide_results:
                    guide_results[best_align.sm.target] = 0
                guide_results[best_align.sm.target] += query_dict[original_seq]
                mapped += query_dict[original_seq]

    logging.info(f"Mapped: {mapped}, Multimap: {multimap} , Unmapped: {unmapped}")
    stats.mapped_to_guide_reads = mapped
    stats.multimap_reads = multimap
    stats.unmapped_reads = unmapped
    stats.total_guides = len(library.guides)
    return (query_dict, guide_results, aligned_results, stats)



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import gzip
import re
from typing import Dict
from typing import List
from typing import TextIO
from typing import Tuple

import yaml
from pkg_resources import resource_string

from pycroquet.classes import Guide
from pycroquet.classes import Library
from pycroquet.classes import LibraryHeader
from pycroquet.targets import guides_to_targets

"""
See: https://docs.google.com/document/d/18G5kT0Z6gmrUwEq6JZ-sd0XuZf7iG5WsDLBY2Bb_uLE/edit

This document needs pulling into the README.

Flexible file definition, some core fields and then additional columns... but that means we need a dynamic parser.
"""

H_INFO_RE = re.compile("^##([^:]+): ?(.+)$")
ACGT_ONLY = re.compile("[ACGT]+")
LIBRARY_TYPES = ("single", "dual", "other")
LIBRARY_STRAND_DEFAULTS = {"single": (0), "dual": (0, 1)}


def _load_config():
    return yaml.safe_load(resource_string(__name__, f"resources/library.yaml").decode("utf-8", "strict"))


def parse_header(ifh: TextIO, line: str) -> LibraryHeader:
    """
    Expects file handle, and first line of file (to cope with open vs gz.open).
    Capture any info header items '##', validate any we decide to be required.
    Finally pass the column '#' row for parsing, do not read another line following this.
    """

    config = _load_config()

    info_items = {}
    lib_type = None
    reverse_read_order = False
    while line and line.startswith("##"):
        # these are the info lines
        line = line.rstrip()
        m_res = H_INFO_RE.match(line)
        if m_res is None:
            raise ValueError("Header info line is not of expected format '^##([^:]+): ?(.+)$': " + line)
        (tag, value) = m_res.group(1, 2)
        if tag in info_items:
            raise ValueError(f"Duplicate key '{tag}' found for '##' header line")
        if tag == "dual-orientation" and value == "R2_R1":
            reverse_read_order = True
        if tag == "library-type":
            if value not in LIBRARY_TYPES:
                raise ValueError(f"Value for 'library-type' ({value}) is not valid, choose from: " + str(LIBRARY_TYPES))
            lib_type = value
        info_items[tag] = value
        line = ifh.readline()
    (cols, col_idx_map, col_sep) = columns_from_header(config, line, lib_type)
    if "library-type" not in info_items:
        info_items["library-type"] = "UNKNOWN"
    return LibraryHeader(
        column_map=col_idx_map,
        column_list=cols,
        info_items=info_items,
        required_cols=config["columns"]["required"],
        config=config,
        column_separators=col_sep,
        reverse_read_order=reverse_read_order,
    )


def columns_from_header(config, line: str, lib_type: str = None) -> Tuple[List, Dict[int, str], Dict[str, str]]:
    """
    this should have single leading #
    """
    line = line.rstrip()
    if not line.startswith("#"):
        raise ValueError("Failed to find the column headers in the guide library file")
    # map columns to index positions so easy to map data into object/dicts
    idx_to_col = {}
    # also capture the list of headings seen
    cols = []
    # also remove any left whitespace after the # just to minimise errors
    for i, c in enumerate(line.lstrip("#").lstrip().split("\t")):
        idx_to_col[i] = c
        cols.append(c)

    conf_cols = config["columns"]
    required_keys = list(conf_cols["required"].keys())
    optional_keys = list(conf_cols["optional"].keys())

    for required in required_keys:
        if required not in cols:
            raise ValueError(f"Required column ({required}) missing from column headers")

    for c in cols:
        if c not in required_keys + optional_keys:
            raise ValueError(f"Column ({c}) is not an expected column")

    if "sgrna_strands" not in cols and (lib_type is None or lib_type not in ("single", "dual")):
        raise ValueError("Either column 'sgrna_strands' or info header '##library-type' (single/dual) must be defined")

    split_cols = {}
    for c in cols:
        if c in required_keys and conf_cols["required"][c] is not None and "separator" in conf_cols["required"][c]:
            split_cols[c] = conf_cols["required"][c]["separator"]
            continue
        if c in optional_keys and conf_cols["optional"][c] is not None and "separator" in conf_cols["optional"][c]:
            split_cols[c] = conf_cols["optional"][c]["separator"]

    return (cols, idx_to_col, split_cols)


def _validate_strands(lib_type, this_no, strand_no, line):
    if lib_type != "UNKNOWN":
        raise ValueError(
            f"'sgrna_strands' column elements ({this_no}) mismatch vs rule for '##library-type: {lib_type}' ({strand_no})"
        )
    else:
        raise ValueError(
            f"'sgrna_strands' column elements ({this_no}) mismatch against previous rows ({strand_no}), line: {line}"
        )


def _strand_info(lh: LibraryHeader) -> Tuple[int, Tuple[int]]:
    strand_no = None
    strand_set = None
    if lh.info_items["library-type"] == "single":
        strand_no = 1
        strand_set = LIBRARY_STRAND_DEFAULTS[lh.info_items["library-type"]]
    elif lh.info_items["library-type"] == "dual":
        strand_no = 2
        strand_set = LIBRARY_STRAND_DEFAULTS[lh.info_items["library-type"]]
    return (strand_no, strand_set)


def parse_data_rows(lh: LibraryHeader, ifh: TextIO) -> List[Guide]:
    col_list = lh.column_list
    col_no = len(col_list)
    col_map = lh.column_map
    required_cols = lh.required_cols
    (strand_no, strand_set) = _strand_info(lh)
    idx = 0
    guides = []

    cols_to_split = lh.column_separators

    while True:
        line = ifh.readline()
        if line == "":
            break
        elements = line.strip().split("\t")
        if len(elements) != col_no:
            raise ValueError(
                f"Column header indicates {col_no} columns, but record has only {len(elements)} (all fields require a value, '.' to omit)\n> {line}"
            )
        guide = Guide(idx=idx)
        for pos, key in col_map.items():
            value = elements[pos]
            if key in cols_to_split:
                # autosplit
                value = value.split(cols_to_split[key])
            if key in required_cols or key == "sgrna_strands":
                setattr(guide, key, value)
            else:
                guide.other[key] = value

        if guide.sgrna_strands is None:
            guide.sgrna_strands = strand_set
        else:
            this_no = len(guide.sgrna_strands)
            if strand_no is None:
                strand_no = this_no
            elif strand_no != this_no:
                _validate_strands(lh.info_items["library-type"], this_no, strand_no, line)

        expect_len = None
        initial_col = None
        for col in cols_to_split:
            if col in guide.other:
                split_len = len(guide.other[col])
            else:
                split_len = len(getattr(guide, col))
            if expect_len is None:
                expect_len = split_len
                initial_col = col
            elif expect_len != split_len:
                raise ValueError(
                    f"All columns that can be split should have the same number of elements.  Column {col} ({split_len}), differs from {initial_col} ({expect_len})"
                )

        if lh.reverse_read_order:
            guide.sgrna_seqs = list(reversed(guide.sgrna_seqs))
        for seq in guide.sgrna_seqs:
            if ACGT_ONLY.fullmatch(seq) is None:
                print(guide)
                raise ValueError(
                    f"'sgrna_seqs' can only contain ACGT and the separator character '|'.  Got '{seq}' after splitting"
                )

        guides.append(guide)
        idx += 1

    if strand_no == 1 and lh.info_items["library-type"] == "UNKNOWN":
        lh.info_items["library-type"] = "single"

    if lh.info_items["library-type"] == "single":
        lh.is_single = True

    return guides


def load(library_file: str) -> Library:
    i_fh = None
    line = None
    try:
        i_fh = gzip.open(library_file, "rt")
        line = i_fh.readline()
    except gzip.BadGzipFile:
        i_fh = open(library_file, "rt")
        line = i_fh.readline()
    lh = parse_header(i_fh, line)
    guides = parse_data_rows(lh, i_fh)
    i_fh.close()

    (target_to_guides, uniq_targets) = guides_to_targets(guides, lh.is_single)
    return Library(
        header=lh,
        guides=guides,
        targets=uniq_targets,
        target_to_guides=target_to_guides,
    )



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import sys

from pycroquet import cli
from pycroquet import countwriter
from pycroquet import libparser
from pycroquet import main
from pycroquet import readparser
from pycroquet import readwriter


def run(
    guidelib,
    queries,
    sample,
    output,
    workspace,
    rules,
    low_count,
    minscore,
    qual_offset,
    cpus,
    reference,
    excludeqcf,
    boundary_mode,
    unique_only,
    chunks,
    no_alignment,
    loglevel,
):
    (usable_cpu, work_tmp, workspace, boundary_mode) = cli.common_setup(
        loglevel, cpus, workspace, output, boundary_mode
    )

    reverse = False
    if unique_only is True:
        (_, stats, query_dict, _) = readparser.parse_reads(
            queries, sample=sample, cpus=cpus, reference=reference, exclude_qcfail=excludeqcf
        )
        countwriter.query_counts(query_dict, stats, output)

    if unique_only is False:
        library = libparser.load(guidelib)

        min_target_len = library.min_target_len()
        minscore = min_target_len - 10

        (query_dict, guide_results, aligned_results, stats) = main.process_reads(
            library,
            queries,
            workspace,
            rules,
            minscore,
            usable_cpu,
            chunks,
            sample=sample,
            reference=reference,
            exclude_qcfail=excludeqcf,
            reverse=reverse,
            exclude_by_len=min_target_len,
            boundary_mode=boundary_mode,
        )
        countwriter.query_counts(query_dict, stats, output)
        countwriter.guide_counts_single(library, guide_results, output, stats, low_count)
        if no_alignment is False:
            readwriter.reads_to_hts(
                library,
                aligned_results,
                queries,
                qual_offset,
                workspace,
                stats,
                output,
                usable_cpu,
                exclude_qcfail=excludeqcf,
                reference=reference,
                reverse=reverse,
            )

    # TODO: write everything out and then if mapped count is a very low fraction repeat.  If the revcomp result is a higher
    # fraction then replace data

    if not work_tmp:
        shutil.rmtree(workspace)



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import gzip
import logging
import os
import re
from time import time
from typing import Dict
from typing import TextIO
from typing import Tuple

import magic
import pysam
from pygas.matrix import revcomp

from pycroquet.classes import Stats
from pycroquet.constants import EXT_TO_HTS
from pycroquet.htscomm import hts_reader


ILLUMINA_SINGLE_FASTQ_HEADER_PATTERN = re.compile(r"^@([^\s/]+)$")
ILLUMINA_FASTQ_HEADER_PATTERN = re.compile(r"^@(\S+)/([12])$")
CASAVA_FASTQ_HEADER_PATTERN = re.compile(r"^@(\S+)\s([012]):([YN])+:[\d+]+:\S+$")

LOAD_INFO_THRESHOLD = 1000000
OFFSET_ILLUMINA = 64
OFFSET_CASAVA = 33


def parse_fq_header(header: str):
    qc_fail = False
    phred_offset = OFFSET_ILLUMINA
    if (match := ILLUMINA_SINGLE_FASTQ_HEADER_PATTERN.match(header)) is not None:
        # illumina format, unparied
        groups = match.groups()
        name = groups[0]
        pair_member = None
    elif (match := ILLUMINA_FASTQ_HEADER_PATTERN.match(header)) is not None:
        # illumina format, paired or single end wit read identifier
        groups = match.groups()
        name = groups[0]
        pair_member = int(groups[1])
    elif (match := CASAVA_FASTQ_HEADER_PATTERN.match(header)) is not None:
        # casava1.8+ format
        phred_offset = OFFSET_CASAVA
        groups = match.groups()
        name = groups[0]
        pair_member = int(groups[1])
        if groups[2] == "Y":
            qc_fail = True
    else:
        raise ValueError(f"Unsupported FastQ header format: {header}")
    return (name, pair_member, qc_fail, phred_offset)


def is_gzip(seq_file):
    magic_types = magic.from_file(seq_file)
    if "gzip compressed data" in magic_types or "gzip compatible" in magic_types:
        return True
    return False


def parse_reads(
    seq_file: str,
    sample,
    cpus,
    reference=None,
    exclude_qcfail=False,
    reverse=False,
    exclude_by_len=None,
    paired=False,
    trim_len=0,
) -> Tuple[int, Stats, Dict[str, int]]:
    """
    This function is for the initial collation of unique read sequences in the original orientation only (hts will do revcomp).
    The is no chunking of data so this relies on large memory lookups at present.

    Selecting correct underlying parser is via file extension:
    - cram/bam/sam -> htslib processing
    - gz assume gzip fastq
    - anything else assume uncompressed fastq

    Returns:
    Tuple[int unique, int total, List[str seq]]
    """
    start = time()
    ext = os.path.splitext(seq_file)[1]

    response = None

    if ext in EXT_TO_HTS:
        logging.info(f"Sequence input detected as *{ext}")
        response = parse_htsfile(
            seq_file,
            EXT_TO_HTS[ext],
            sample,
            cpus,
            reference=reference,
            exclude_qcfail=exclude_qcfail,
            reverse=reverse,
            exclude_by_len=exclude_by_len,
            paired=paired,
            trim_len=trim_len,
        )
    else:
        fq_fh = None
        if is_gzip(seq_file):
            logging.info("Sequence input detected as gzip (assume fastq)")
            fq_fh = gzip.open(seq_file, "rt")
        else:
            logging.info("uncompressed data (assume fastq)")
            fq_fh = open(seq_file)
        response = parse_fastq(
            fq_fh,
            sample,
            exclude_qcfail=exclude_qcfail,
            reverse=reverse,
            exclude_by_len=exclude_by_len,
            paired=paired,
            trim_len=trim_len,
        )
    if paired:
        logging.info(f"Parsed {response[1].total_pairs} pairs, {len(response[3])} were unique...")
    else:
        logging.info(f"Parsed {response[1].total_reads} reads, {response[0]} were unique...")
    logging.info(f"Read parsing took: {int(time() - start)}s")
    return response


def _hts_sample(sam: pysam.AlignmentFile, sample=None):
    if sample is not None:
        return sample
    header = sam.header.as_dict()
    if "RG" not in header:
        return sample
    for rg in header["RG"]:
        if "SM" not in rg:
            continue
        if sample is None:
            sample = rg["SM"]
        elif sample != rg["SM"]:
            raise ValueError("Multiple different sample names found in header")
    return sample


def parse_htsfile(
    seq_file: str,
    mode: str,
    sample,
    cpus,
    reference=None,
    exclude_qcfail=False,
    reverse=False,
    exclude_by_len=None,
    paired=False,
    trim_len=0,
) -> Tuple[int, Stats, Dict[str, int]]:
    sam = hts_reader(seq_file, mode, cpus, reference)

    stats = Stats(sample_name=_hts_sample(sam, sample=sample))
    if stats.sample_name is None:
        raise ValueError("No sample name found in input file header, please provide via '--sample'")

    reads = {}
    (unique, total, len_ex, total_pairs, unique_pairs) = (0, 0, 0, 0, 0)
    last_read = None
    last_seq = None
    pairs = {} if paired else None
    for read in sam.fetch(until_eof=True):
        if read.is_secondary or read.is_supplementary:
            continue
        if read.is_qcfail:
            if exclude_qcfail:
                continue
            # only count them if we keep them
            stats.vendor_failed_reads += 1

        seq = read.get_forward_sequence()
        if trim_len:
            seq = seq[0:trim_len]
        if exclude_by_len and len(seq) < exclude_by_len:
            len_ex += 1
            continue
        if reverse:
            seq = revcomp(seq)
        if paired:
            if not read.is_paired:
                continue
            if read.is_read1:
                last_read = read.query_name
                last_seq = seq
            else:
                if last_read != read.query_name:
                    raise ValueError("Paired reads require collation before parsing")
                pair_seq = f"{last_seq}|{seq}"
                total_pairs += 1
                if pair_seq in pairs:
                    pairs[pair_seq] += 1
                else:
                    pairs[pair_seq] = 1
                    unique_pairs += 1
        if seq in reads:
            reads[seq] += 1
        else:
            reads[seq] = 1
            unique += 1
        total += 1
        if total % LOAD_INFO_THRESHOLD == 0:  # pragma: no cover
            if paired:
                logging.debug(f"Parsed {total} reads, {unique} were unique...")
            else:
                logging.debug(f"Parsed {total_pairs} pairs, {unique_pairs} were unique...")
    sam.close()
    stats.total_reads = total
    stats.total_pairs = total_pairs
    stats.reversed_reads = reverse
    if exclude_by_len:
        stats.length_excluded_reads = len_ex
    return (unique, stats, reads, pairs)


def parse_fastq(
    ifh: TextIO,
    sample,
    exclude_qcfail=False,
    reverse=False,
    exclude_by_len=None,
    paired=False,
    trim_len=0,
) -> Tuple[int, Stats, Dict[str, int]]:
    """
    Closes received file handle
    Only used for the reads seq minimization process
    """
    stats = Stats(sample_name=sample)
    if stats.sample_name is None:
        raise ValueError("--sample must be provided for fastq inputs")

    if paired:
        raise ValueError("fastq doesn't support paired processing yet")

    reads = {}
    (unique, total, len_ex) = (0, 0, 0)
    pairs = {} if paired else None
    header = ifh.readline()
    while header:
        seq = ifh.readline().strip()
        _ = ifh.readline()  # throw away separator
        _ = ifh.readline()  # throw away qual at this point
        (_, _, qc_fail, _) = parse_fq_header(header.strip())  # for validation only here
        # looks odd, but best way to handle end of file
        header = ifh.readline()
        if qc_fail:
            if exclude_qcfail:
                continue
            # only count them if we keep them
            stats.vendor_failed_reads += 1
        if trim_len:
            seq = seq[0:trim_len]
        if exclude_by_len and len(seq) < exclude_by_len:
            len_ex += 1
            continue
        if reverse:
            seq = revcomp(seq)
        if seq in reads:
            reads[seq] += 1
        else:
            reads[seq] = 1
            unique += 1
        total += 1
        if total % LOAD_INFO_THRESHOLD == 0:  # pragma: no cover
            logging.debug(f"Parsed {total} reads, {unique} were unique...")
    ifh.close()
    stats.total_reads = total
    stats.reversed_reads = reverse
    if exclude_by_len:
        stats.length_excluded_reads = len_ex
    return (unique, stats, reads, pairs)


def collate(seq_file, workspace, cpus):
    """
    collates reads into pairs IF input is a hts file
    """
    start = time()
    hts_cpus = cpus
    if cpus > 4:
        hts_cpus = 4
    ext = os.path.splitext(seq_file)[1]
    if ext not in EXT_TO_HTS:
        return seq_file
    tmp_hts = os.path.join(workspace, "collated.reads.bam")
    logging.info(f"Collating pairs from {seq_file} to {tmp_hts}")
    collate_opts = ["-f", "-l", "1", "--no-PG", "-@", str(hts_cpus), "-o", tmp_hts, seq_file]
    logging.debug("samtools collate " + " ".join(collate_opts))
    pysam.collate(*collate_opts)
    logging.info(f"Collation of reads took: {int(time() - start)}s")
    return tmp_hts



#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
import logging
from typing import Dict
from typing import List
from typing import Tuple

from pygas.matrix import revcomp

from pycroquet.classes import Guide

"""
Handles minimizing target sequences and the de-convolution back to the guide sets they came from
Overall this is is easy for singe guide but more complex for multi-guide
"""


def guides_to_targets(guides: List[Guide], single: bool = False) -> Tuple[Dict[str, List[int]], List[str]]:
    # this links a target sequence to the guide library position(s) in the list
    # target orientation is that of first added, we then check for equivalent targets
    target_to_guides = {}  # Dict[str, List[int]]
    total_dups = 0
    for g_idx, guide in enumerate(guides):
        for target in guide.sgrna_seqs:
            if target in target_to_guides:
                if single:
                    logging.debug(f"Guide duplicate: {guide.id}")
                    total_dups += 1
                    # this only works for single guide
                    guide.unique = False
                    guides[target_to_guides[target][0]].unique = False
                target_to_guides[target].append(g_idx)
            else:
                target_to_guides[target] = [g_idx]

    # now we need the unique list of targets, in a stable order
    uniq_targets = list(target_to_guides.keys())
    uniq_targets.sort()
    if single:
        logging.info(f"Number of duplicate guides: {total_dups}")
    logging.info(f"Total unique guides: {len(uniq_targets)}")
    return (target_to_guides, uniq_targets)



#!/usr/bin/env python3
#
# Copyright (c) 2021
#
# Author: CASM/Cancer IT <cgphelp@sanger.ac.uk>
#
# This file is part of pycroquet.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# 1. The usage of a range of years within a copyright statement contained within
# this distribution should be interpreted as being equivalent to a list of years
# including the first and last year specified and all consecutive years between
# them. For example, a copyright statement that reads ‘Copyright (c) 2005, 2007-
# 2009, 2011-2012’ should be interpreted as being identical to a statement that
# reads ‘Copyright (c) 2005, 2007, 2008, 2009, 2011, 2012’ and a copyright
# statement that reads ‘Copyright (c) 2005-2012’ should be interpreted as being
# identical to a statement that reads ‘Copyright (c) 2005, 2006, 2007, 2008,
# 2009, 2010, 2011, 2012’.
from setuptools import setup

config = {
    "name": "pycroquet",
    "description": "python Crispr Read to Oligo QUantification Enhancement Tool",
    "long_description": open("README.md").read(),
    "long_description_content_type": "text/markdown",
    "author": "Keiran M Raine",
    "url": "https://github.com/cancerit/pycroquet",
    "author_email": "cgphelp@sanger.ac.uk",
    "version": "1.4.0",
    "license": "AGPL-3.0",
    "python_requires": ">= 3.9",
    "install_requires": ["click", "click-option-group", "python-magic", "pysam", "pygas", "PyYAML"],
    "packages": ["pycroquet"],
    "setup_requires": ["click"],
    "test_suite": "tests",
    "tests_require": ["pytest"],
    "entry_points": {
        "console_scripts": ["pycroquet=pycroquet.cli:cli"],
    },
    "package_data": {
        "pycroquet": [
            "resources/library.yaml",
        ],
    },
}

setup(**config)
